// To compile
// g++ -lming -lboost_regex -o FreeFem2Swf FreeFem2Swf.cpp 
// 
// To use
// FreeFem2Swf [-bb=minx,miny,max,may -dim=nx,ny] output.swf [Th.msh]*
//
// bb = bounding box
// dim= dimension of the flash
//
//  To use on all the file Th.%d.msh
//  ls *.msh | sort -k2 -th -n | xargs FreeFem2Swf -bb=-2.5,-1,3,5 test.swf 
//   (for n in {10..44}; do echo Th$n.msh; done) | xargs FreeFem2Swf -bb=-2.5,-1,3,5 test.swf 
//
//  BUG: When to many (or too big ?) files are given in arguments, the behavior became weird at a point... don' know why
//
// on one example:
// 8377767 compression level=1 .
// 8139178 compression level=9

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <string>
//#include <mingpp.h>	//Ming cpp library for creating swf files 
#include <mingpp.h>	//Ming cpp library for creating swf files 
#include "vector"
#include <math.h>
#include <boost/regex.hpp> // I tried <regex> but it didn't seem to work properly.

using namespace std;

double const pi=4.*atan(1.);

class vertex {
	private:
		double x1,x2;
		size_t label;
	public:
		vertex();
		vertex(double const&,double const&,size_t const&);
		friend class mesh;
		friend class triangle;
		friend class edge;
		friend class flash;
};
class triangle {
	private:
		size_t a1,a2,a3,label;
	public:
		triangle();
		triangle(size_t const&,size_t const&,size_t const&,size_t const&);
		friend class mesh;
		friend class flash;
};
class edge {
	private:
		size_t e1,e2,label;
	public:
		edge();
		edge(size_t const&,size_t const&,size_t const&);
		friend class mesh;
		friend class flash;
};
class mesh {
	private:
		size_t nv,nt,ne;
		vector<vertex> x;	
		vector<triangle> K;
		vector<edge> e;
	public:
		mesh(const string);
	friend class flash;
};
class EF {
	private:
		mesh const* Mesh_;
	public:
		double umax,umin;		// il faudrait plutot les mettre en prive... de sorte a ne pas pouvoir les modifier
		vector<double> u;		// idem ici, acces libre... mais mise a jour necessite de rafraichir umax et umin
		EF(mesh const&,char const* const);
	friend class flash;
};
class flash {
	private:
		SWFMovie *Movie;
		vector<SWFDisplayItem*> items;
		double infX,infY,supX,supY;
		size_t dimX,dimY;
		double coef,deltaX,deltaY;
	public:
		void add(EF const&);
		void add(mesh const&,EF const* const=NULL);
		bool save(const string);
		void next();
		flash(size_t const&,size_t const&,double* const&);
};


////////////////////////////////////////////////////////////////////////////
	flash::flash(size_t const& dimX_,size_t const& dimY_,double* const& tab):Movie(new SWFMovie),infX(tab[0]),infY(tab[1]),supX(tab[2]),supY(tab[3]),dimX(dimX_),dimY(dimY_){
				Movie->setDimension(dimX,dimY);
				double fps=4;
				Movie->setRate(fps);							// Set the nb of frames per second		
				//Movie->setFrames(19); 							// Number of frames
				// Compute rescaling variables
				{
					double coefX,coefY;
					coefX=dimX/(supX-infX);
					coefY=dimY/(supY-infY);
					coef=min(coefX,coefY);
					double DimX=coef*(supX-infX);
					double DimY=coef*(supY-infY);
					deltaX=(dimX-DimX)/2.;
					deltaY=(dimY-DimY)/2.;
				}
		};
	void flash::add(mesh const& Mesh,EF const* const u){
			// We add the inner mesh
			{
				for(size_t i=0;i<Mesh.nt;i++){
						triangle const& T=Mesh.K[i];	
					// Le style devrait etre pouvoir etre adapte a souhait.........
						double X[3]={Mesh.x[T.a1].x1,Mesh.x[T.a2].x1,Mesh.x[T.a3].x1};
						double Y[3]={Mesh.x[T.a1].x2,Mesh.x[T.a2].x2,Mesh.x[T.a3].x2};
						for(int i=0;i<3;i++){
							X[i]=(X[i]-infX)*coef+deltaX;
							Y[i]=(supY-Y[i])*coef+deltaY;
						}
						double v[3]={1.,1.,1.}; // 1=noir, 0=blanc
						if (u!=NULL)
						{
								v[0]=u->u[T.a1];v[1]=u->u[T.a2];v[2]=u->u[T.a3];
								for(int i=0;i<3;i++)
									v[i]=(v[i]-u->umin)/(u->umax-u->umin);
						}
						if ((X[1]-X[0])*(Y[2]-Y[0])-(X[2]-X[0])*(Y[1]-Y[0])<0){
							double inter;
							inter=X[1];X[1]=X[2];X[2]=inter;
							inter=Y[1];Y[1]=Y[2];Y[2]=inter;
							inter=v[1];v[1]=v[2];v[2]=inter;
						}
						
						SWFShape* shape=new SWFShape;		
						
						double Xmin=min(X[0],min(X[1],X[2])),Ymin=min(Y[0],min(Y[1],Y[2]));
						double Xmax=max(X[0],max(X[1],X[2])),Ymax=max(Y[0],max(Y[1],Y[2]));						
						double dudx= (Y[2]-Y[0])*(v[1]-v[0])-(Y[1]-Y[0])*(v[2]-v[0]);
						double dudy=-(X[2]-X[0])*(v[1]-v[0])+(X[1]-X[0])*(v[2]-v[0]);
						double theta=atan2(dudy,dudx);
						double thetas=atan2(sin(theta)*(Ymax-Ymin),cos(theta)*(Xmax-Xmin));
						
						double cMin=255*min(v[0],min(v[1],v[2]));
						double cMax=255*max(v[0],max(v[1],v[2]));

						double rhoMin,rhoMax;
						{
							double RMin=min(min(Xmin*cos(theta)+Ymin*sin(theta),
							Xmin*cos(theta)+Ymax*sin(theta)),min(
							Xmax*cos(theta)+Ymin*sin(theta),
							Xmax*cos(theta)+Ymax*sin(theta)));
							double RMax=max(max(
							Xmin*cos(theta)+Ymin*sin(theta),
							Xmin*cos(theta)+Ymax*sin(theta)),
							max(Xmax*cos(theta)+Ymin*sin(theta),Xmax*cos(theta)+Ymax*sin(theta)));
							double rmin=min(X[0]*cos(theta)+Y[0]*sin(theta),
							min(X[1]*cos(theta)+Y[1]*sin(theta),
							X[2]*cos(theta)+Y[2]*sin(theta)));
							double rmax=max(
							X[0]*cos(theta)+Y[0]*sin(theta),
							max(X[1]*cos(theta)+Y[1]*sin(theta),X[2]*cos(theta)+Y[2]*sin(theta)));
							rhoMin=(rmin-RMin)/(RMax-RMin);
							rhoMax=(rmax-RMin)/(RMax-RMin);
							
						}	

						SWFGradient *gradient = new SWFGradient();
						/*
						double dt=0.5;
						for(double t=0;t<=1;t+=dt){
							size_t c=cMax+(cMin-cMax)*t;
							double rho=rhoMin+(rhoMax-rhoMin)*t;
							gradient->addEntry(rho,c,c,255);
							//gradient->addEntry(rhoMin,cMax,255,255); // de maniere etrange, cela ne fonctionne pas avec a...
							//gradient->addEntry(rhoMax,cMin,255,255); // de maniere etrange, cela ne fonctionne pas avec a...
						}
						*/
						
						
						gradient->addEntry(rhoMin,cMin,cMin,cMin);//// en fait, cela ne semble fonctionner que pour ce choix !!!
						gradient->addEntry(rhoMax,cMax,cMax,cMax);// // de maniere etrange, cela ne fonctionne pas avec a...
						
						//gradient->addEntry(1.,0,255,255);// // de maniere etrange, cela ne fonctionne pas avec a...
						
						
						//gradient->addEntry(rhomin,0,cMin,255);//// en fait, cela ne semble fonctionner que pour ce choix !!!
						//gradient->addEntry(1,0,255,255);// // de maniere etrange, cela ne fonctionne pas avec a...
						
						SWFFill* fill=shape->addGradientFill(gradient, SWFFILL_LINEAR_GRADIENT);
						//fill->scaleTo(0.1,1.); // plus la valeur est gradne, plus de gradient est plat..
						//fill->moveTo(0.,0.);
						fill->rotate(-thetas*360/(2.*pi));	// Ok, le gradient est oriente dans la bonne direction...
						shape->setRightFill(fill);	// A mettre avant de tracer la forme
						
						shape->setLine(1,100,100,100);
						//shape->hideLine();
						shape->movePenTo(X[0],Y[0]);
						shape->drawLineTo(X[1],Y[1]);
						shape->drawLineTo(X[2],Y[2]);
						shape->drawLineTo(X[0],Y[0]);
						
						items.push_back(Movie->add(shape));
						// remove shape, gradient and fill pt
					
			/*
			{
				SWFShape* shape=new SWFShape;
				shape->setLine(1,0,255,0);
				shape->movePenTo((X[0]+X[1]+X[2])/3.,(Y[0]+Y[1]+Y[2])/3.);
				shape->drawLine(cos(theta)*10,sin(theta)*10);
				shape->setLine(1,255,0,0);
				shape->drawLine(cos(theta)*8,sin(theta)*8);
				items.push_back(Movie->add(shape));
			}*/
					//delete shape; // Why doesn't he want's me to remove the shape ??? It generates an error
				}
			}
			// We add the boundary
			{
				SWFShape *shape=new SWFShape();
				
				// Le style devrait etre pouvoir etre adapte a souhait.........
				shape->setLine(2,255,0,0); 
				
				for(size_t i=0;i<Mesh.ne;i++)
				{
					edge const& b=Mesh.e[i];	
					{
						double const& x=Mesh.x[b.e1].x1;;
						double const& y=Mesh.x[b.e1].x2;
						shape->movePenTo((x-infX)*coef+deltaX,(supY-y)*coef+deltaY);
					}
					{
						double const& x=Mesh.x[b.e2].x1;;
						double const& y=Mesh.x[b.e2].x2;
						shape->drawLineTo((x-infX)*coef+deltaX,(supY-y)*coef+deltaY);
					}
				}
				items.push_back(Movie->add(shape));
				//delete shape; // Why doesn't he want's me to remove the shape ???
			}
};
	void flash::add(EF const& u){
		this->add(*(u.Mesh_),&u);
	}
	void flash::next(){	
		// go to next frame
		Movie->nextFrame();
		// remove all the items
		size_t iMax=items.size();
			for(size_t i=0;i<iMax;i++) items[i]->remove();
		items.resize(0);
	};
	bool flash::save(const string Name_File){
		if (items.size()!=0) next();
		Movie->save(Name_File.c_str(),1);		// Save the swf movie file to disk. Compression level set to 1-9
		return true;
	};
	// Vertex, triangle, edege classes //
	vertex::vertex(double const& x1_,double const& x2_,size_t const& label_):x1(x1_),x2(x2_),label(label_){};
	vertex::vertex(){};
	triangle::triangle(size_t const& a1_,size_t const& a2_,size_t const& a3_,size_t const& label_):a1(a1_),a2(a2_),a3(a3_),label(label_){};
	triangle::triangle(){};
	edge::edge(size_t const& e1_,size_t const& e2_,size_t const& label_):e1(e1_),e2(e2_),label(label_){};
	edge::edge(){};
	
	// Mesh class //
	mesh::mesh(const string FileName){
		ifstream Read(FileName.c_str()); // c_str convert string 2 char*
		if(Read.is_open()){
		Read>>nv;Read>>nt;Read>>ne;
		// Get the vertices
		{
			x.resize(nv);
			double x1,x2; size_t label;
			for(size_t i=0;i<nv;i++){
				Read>>x1;Read>>x2;Read>>label;
				x[i]=vertex(x1,x2,label);
			}
		}
		// Get the triangles
		{
			K.resize(nt);
			size_t a1,a2,a3,label;
			for(size_t i=0;i<nt;i++){
				Read>>a1;Read>>a2;Read>>a3;Read>>label;
				K[i]=triangle(a1-1,a2-1,a3-1,label);
			}
		}
		
		// Get the boundaries
		{
			e.resize(ne);
			size_t e1,e2,label;
			for(size_t i=0;i<ne;i++){
				Read>>e1;Read>>e2;Read>>label;
				e[i]=edge(e1-1,e2-1,label);
			}
		}}
		else {
			cout<<"Could not open file "<<FileName<<endl;
			exit(EXIT_FAILURE);
		}
	};

	// Finite Element class //
	EF::EF(mesh const& Mesh,char const* const FileName):Mesh_(&Mesh){
		ifstream Read(FileName);
		size_t sizeU;
		if(Read.is_open()){
			if (!(Read>>sizeU)){
				cout<<"File "<<FileName<<" is empty"<<endl;exit(EXIT_FAILURE);
			} else {
				double U;
				Read>>U;
				u.resize(0);
				u.push_back(U);
				umax=U;
				umin=umax;
				while(Read>>U){
					u.push_back(U);
					umin=min(umin,U);umax=max(umax,U);
				}
			}
		} else {
			cout<<"Could not open file "<<FileName<<endl;
			exit(EXIT_FAILURE);
		}
	}
//////////////////                  MAIN                  ////////////////////

void ArgumentsError()
{
		cout<<"FreeFem2Swf -- version 1 --- 27/03/2014"<<endl;
		cout<<"-- Use --------"<<endl;
		cout<<"     FreeFem2Swf -bb=[a,b,c,d] -size=dX,dY output.swf file.msh [file.msh]* "<<endl;
		cout<<"            output.swf          = Output file (swf format)"<<endl;
		cout<<"            file.msh            = Mesh files (msh format)"<<endl;
		cout<<endl;
		cout<<"Incorrect Number of arguments or format"<<endl;
		exit(1);
}

int main(int argc,char* argv[])
{
	Ming_init();	
	size_t dimX=600,dimY=600;	// default size of the flash
	double bb[4]={0,0,1,1};		// default bounding box

	int currentArgument=1;
	if(argc==1) ArgumentsError();
	while (argv[currentArgument][0]=='-')
	{
		// It is an option
		bool found=false;	// can it be found ?
		{// bounding box definition
			boost::regex e ("-bb=(-?[0-9\\.]*),(-?[0-9\\.]*),(-?[0-9\\.]*),(-?[0-9\\.]*)");
			boost::match_results<std::string::const_iterator>  sm;
			if(boost::regex_match ((string) argv[currentArgument],sm,e))
			{	
				string l0=sm[1];
				string l1=sm[2];
				string l2=sm[3];
				string l3=sm[4];
				bb[0]=atoi(l0.c_str());
				bb[1]=atoi(l1.c_str());
				bb[2]=atoi(l2.c_str());
				bb[3]=atoi(l3.c_str());
				currentArgument++;found=true;
			}
		}
		{// Dimension of the flash
			boost::regex e("-dim=([0-9]*),([0-9]*)");
			boost::match_results<std::string::const_iterator>  sm;
			if(boost::regex_match ((string) argv[currentArgument],sm,e))
			{	
				string X=sm[1];
				string Y=sm[2];
				dimX=atoi(X.c_str());
				dimY=atoi(Y.c_str());
				currentArgument++;found=true;
			}
		}
		if(!found)
		{
			cout<<"Option "<<argv[currentArgument]<<"Not Found"<<endl;exit(1);
		}
	}
	// read output file
	string output;
	output=argv[currentArgument++];
	{// check if *.swf
		boost::regex e(".*\\.swf");
		if (!(boost::regex_match (output,e)))
		{
			cout<<"Output file should have swf extension\n";
			exit(EXIT_FAILURE);
		}
	}

	flash movie(dimX,dimY,bb);
	while(currentArgument<argc){
		//cout<<"i="<<i<<endl;
		// We read the mesh and P1-finite elements files //
		string meshFile=argv[currentArgument++];	
		cout<<"file="<<meshFile<<endl;
		// We add the finite element to the movie
		mesh Sh(meshFile);
		movie.add(Sh);
		// go to next frame
		movie.next();
	}
	// We save the movie
	movie.next(); // clear all ?
	movie.save(output);
	return EXIT_SUCCESS;
}
//////////////////////////////////////////////////////////////////////////////
